/*****************************************************************************
 *
 * Copyright (c) 2020, Marcel Schoppmeier
 * Westfälische Wilhelms-Universität Münster, AG Embedded Systems 
 * All rights reserved.
 *
 * This file is part of SysCSecT
 *
 * SysCSecT is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free
 * Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 *
 * SysCSecT is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SysCSecT.  If not, see <http://www.gnu.org/licenses/>.
 *
 *
 *  Please report any problems or bugs to: sysc.sec@uni-muenster.de
 *
 ****************************************************************************/


#ifndef RSA_CPP
#define RSA_CPP

#include "Utils.cpp"
#include "constants.h"

/* 
 * Modulo function
 * 
 * @param x Number
 * @param n Modulus
 * @return result
 */
inline bigint modulo(bigint x, bigint n)
{
    if(x >= n)
    {
        wait(MOD, SC_PS);
        x = x % n;
    }
    return x;
}

/* 
 * Calculates x*y with the booth algorithm
 * 
 * @param x First factor
 * @param y Second factor
 * @return result
 */
inline bigint produceBoothMulti(const bigint& x, const bigint& y)
{
    bigint result = 0;
    bigint r = x;
    
    int timeSub1 = 0;
    int timeSub2 = 0;
    int timeAdd1 = 0;
    int timeAdd2 = 0;
    
    int b1 = 0;
    int b2 = y[0];
    
    if (b2 == 1)
    {
        result -= r;
        
        timeSub1 = Utils::numberBits(result);
        timeSub2 = Utils::numberBits(r);
        wait(timeSub1 + timeSub2, SC_PS);
        wait(ASSIGN,SC_PS);
       
    }
    r <<= 1;
    wait(SHIFT, SC_PS);
    
    for(int i = 1; i < MAXBITS; ++i)
    {
        b1 = y[i-1];
        b2 = y[i];
        
        if( b2 == 0 && b1 == 1)
        {
            result += r;
            
            timeAdd1 = Utils::numberBits(result);
            timeAdd2 = Utils::numberBits(r);
            wait(timeAdd1 + timeAdd2, SC_PS);
            wait(ASSIGN,SC_PS);
        }
        else if (b2 == 1 && b1 == 0)
        {
            result -=r;
            
            timeSub1 = Utils::numberBits(result);
            timeSub2 = Utils::numberBits(r);
            wait(timeSub1 + timeSub2, SC_PS);
            wait(ASSIGN,SC_PS);
        }
        
        r <<= 1;
        wait(SHIFT, SC_PS);
    }
    return result;
}

/* 
 * the square-and-multiply algorith for the calculation of s = x^d % n
 * 
 * @param x Message
 * @return s
 */
inline bigint squaringAlgorithm(const bigint& x)
{
    bigint d = dValue;//private key
    bigint n = VALUE_N;//Modulo
    bigint s = x;
    
    int timeS = 0;
    int timeX = 0;
    
    //mulitplication effort depends only on highestSigBit(x) and not on the intermediate values s
    if(VARIANT == 1)
    {
        
        timeX = Utils::highestSigBit(x);
        
        for(int i = Utils::highestSigBit(d) - 2; i >= 0; --i)
        {
            s = (s * s) % n;
            wait(timeX,SC_PS);
            wait(ASSIGN,SC_PS);

            if(d[i] == 1)
            {
               s = (s * x) % n;
               wait(timeX,SC_PS);
               wait(ASSIGN,SC_PS);
            }
        }
    }
    //mulitplication effort depends on the highest significant bit also for the intermediate values s
    else if(VARIANT == 2)
    {
        for(int i = Utils::highestSigBit(d) - 2; i >= 0; --i)
        {
            timeS = Utils::highestSigBit(s);
            if((s * s) < n)
            {
                s = s * s;
                wait(timeS * 2,SC_PS);
            }
            else
            {
                s = (s * s) % n;
                wait(timeS * 2,SC_PS);
                wait(MOD,SC_PS);
            }
            wait(ASSIGN,SC_PS);
            
            if(d[i] == 1)
            {
                timeX= Utils::highestSigBit(x);
                timeS = Utils::highestSigBit(s);
                
                if((s * x) < n)
                {
                    s = s * x;
                    wait(timeS + timeX,SC_PS);
                }
                else
                {
                    s = (s * x) % n;
                    wait(timeS + timeX,SC_PS);
                    wait(MOD,SC_PS);
                }
                wait(ASSIGN,SC_PS);
            }
        }
    }
    //mulitplication effort depends on number of bits during calculation
    else if(VARIANT == 3)
    {
        for(int i = Utils::highestSigBit(d) - 2; i >= 0; --i)
        {
            timeS = Utils::numberBits(s);
        
            if((s * s) < n)
            {
                s = s * s;
                wait(timeS * 2, SC_PS);
            }
            else
            {
                s = (s * s) % n;
                wait(timeS * 2, SC_PS);
                wait(MOD, SC_PS);
            }
            wait(ASSIGN,SC_PS);
            
            if(d[i] == 1)
            {
                timeX = Utils::numberBits(x);
                timeS = Utils::numberBits(s);
                
                if((s * x) < n)
                {
                    s = s * x;
                    wait(timeS + timeX, SC_PS);
                } 
                else
                {
                    s = (s * x) % n;
                    wait(timeS + timeX, SC_PS);
                    wait(MOD, SC_PS);
                }
                wait(ASSIGN,SC_PS);
            }
        }
    }
    //mulitplication effort depends on the calculation effort for the Booth algorithm
    else if(VARIANT == 4)
    {
        for(int i = Utils::highestSigBit(d) - 2; i >= 0; --i)
        {
            if((s * s) < n)
            {
                s = produceBoothMulti(s,s);
            }
            else
            {
                wait(MOD,SC_PS);
                s = produceBoothMulti(s,s) % n;
            }
            wait(ASSIGN,SC_PS);
            
            if(d[i] == 1)
            {
                if((s * x) < n)
                {
                    s = produceBoothMulti(s,x);
                }
                else
                {
                    wait(MOD,SC_PS);
                    s = produceBoothMulti(s,x) % n;
                }
                wait(ASSIGN,SC_PS);
            }
        }
    }
    return s;
}
#endif
