/*****************************************************************************
 *
 * Copyright (c) 2020, Marcel Schoppmeier
 * Westfälische Wilhelms-Universität Münster, AG Embedded Systems 
 * All rights reserved.
 *
 * This file is part of SysCSecT
 *
 * SysCSecT is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free
 * Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 *
 * SysCSecT is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SysCSecT.  If not, see <http://www.gnu.org/licenses/>.
 *
 *
 *  Please report any problems or bugs to: sysc.sec@uni-muenster.de
 *
 ****************************************************************************/


#ifndef Simulations_CPP
#define Simulations_CPP

#include "Utils.cpp"
#include "constants.h"

/*
 * Class that encapsulates the simulation which is executed by the timeanalyzer.
 * The class contains the time simulation for one bit guess and two bit guesses.
 */

class Simulations{
    
private:
    
public:
    
    /* 
     * Calculates x*y with the booth algorithm and simulates the time
     * 
     * @param x First factor
     * @param y Second factor
     * @param time Measurement variable
     * @return result
     */
    static bigint produceBoothMultiSimulation(const bigint& x, const bigint& y, int& time)
    {
        bigint result = 0;
        bigint r = x;
        int b1 = 0;
        int b2 = y[0];
        
        int timeSub1 = 0;
        int timeSub2 = 0;
        int timeAdd1 = 0;
        int timeAdd2 = 0;
    
        if (b2 == 1)
        {
            result -= r;
            
            timeSub1 = Utils::numberBits(result);
            timeSub2 = Utils::numberBits(r);
            time += (timeSub1 + timeSub2);
            time += ASSIGN;
        }
        r <<= 1;
        time += SHIFT;
        
        for(int i = 1; i < MAXBITS; ++i)
        {
            b1 = y[i-1];
            b2 = y[i];
            
            if( b2 == 0 && b1 == 1)
            {
                result += r;
                
                timeAdd1 = Utils::numberBits(result);
                timeAdd2 = Utils::numberBits(r);
                time += (timeAdd1 + timeAdd2);
                time += ASSIGN;
            }
            else if (b2 == 1 && b1 == 0)
            {
                result -=r;
                
                timeSub1 = Utils::numberBits(result);
                timeSub2 = Utils::numberBits(r);
                time += (timeSub1 + timeSub2);
                time += ASSIGN;
            }
            
            r <<= 1;
            time += SHIFT;
        }
        return result;
    }
    
    
    
    /* 
     * Two bit guess simulation
     * 
     * @param b1 First bit guess
     * @param b2 Second bit guess
     * @param x  Message
     * @param d  Private key
     * @param guessD[] Bits already guessed
     * @param i Position up to which all bits have already been guessed
     * @return the execution time
     */
    static int calcTimeSimulation_2b(bool b1, bool b2, bigint x, bigint d, bool guessD[], int i)
    {
        bigint n = VALUE_N;//Modulo
        bigint s = x;
        
        int t1 = 0;//variable for the result
        int timeS = 0;//variable for the intermediate time caused by s
        int timeX = 0;//variable for the intermediate time caused by x
        
        if(VARIANT == 2)
        {
            int calcBits = i + 1;//already guessed bit and the first guess bit
            if(i < Utils::highestSigBit(d) - 2)//in last step do not guess the second bit, because there is none
                calcBits += 1;//second bit can also be guessed
            
            //calculation with time simulation
            for(int j = 1; j <= calcBits; j++)
            {
                timeS = Utils::highestSigBit(s);
                
                //s = (s*s)%n; modulo effort is differentiated
                if((s * s) < n)
                {
                    s = s * s;
                    
                    t1 += (timeS * 2);
                }
                else
                {
                    s = (s * s) % n;
                    
                    t1 += (timeS * 2);
                    t1 += MOD;
                }
                t1 += ASSIGN;
                
                //if the bit is set do a multiplication
                if((guessD[j] && j <=i) || (b1 && (j==i+1)) || (b2 && (j == (i+2))) )
                {
                    timeX = Utils::highestSigBit(x);
                    timeS = Utils::highestSigBit(s);
                    
                    //s = (s*x)%n; modulo effort is differentiated
                    if((s * x) < n)
                    {
                        s = s * x;
                        
                        t1 += (timeS + timeX);
                    }
                    else
                    {
                        s = (s * x) % n;
                        
                        t1 += (timeS + timeX);
                        t1 += MOD;
                    }
                    t1 += ASSIGN;
                }
            }
        }else if(VARIANT == 3)
        {
            int calcBits = i + 1;//already guessed bit and the first guess bit
            if(i < Utils::highestSigBit(d) - 2)//in last step do not guess the second bit, because there is none
                calcBits += 1;//second bit can also be guessed
            
            //calculation with time simulation
            for(int j = 1; j <= calcBits; j++)
            {
                timeS = Utils::numberBits(s);
                
                //s = (s*s)%n; modulo effort is differentiated
                if((s * s) < n)
                {
                    s = s * s;
                    
                    t1 += timeS * 2;
                } 
                else
                {
                    s = (s * s) % n;
                    
                    t1 += timeS * 2;
                    t1 += MOD;
                }
                t1 += ASSIGN;
                
                //if the bit is set do a multiplication
                if((guessD[j] && j <= i)||(b1 && (j == (i+1)))||(b2 && (j == (i+2))))
                {
                    timeX = Utils::numberBits(x);
                    timeS = Utils::numberBits(s);
                    
                    //s = (s*x)%n; modulo effort is differentiated
                    if((s * x) < n)
                    {
                        s = s * x;
                        
                        t1 += timeS + timeX;
                    }
                    else
                    {
                        s = (s * x) % n;
                        
                        t1 += timeS + timeX;
                        t1 += MOD;
                    }
                    t1 += ASSIGN;
                }
            }
        }else if(VARIANT == 4)
        {
            int calcBits = i + 1;//already guessed bit and the first guess bit
            if(i < Utils::highestSigBit(d) - 2)//in last step do not guess the second bit, because there is none
                calcBits += 1;//second bit can also be guessed
            
            //calcultion with Booth and time simulation
            for(int j = 1; j <= calcBits; j++)
            {
                //s = (s*s)%n; modulo effort is differentiated
                if((s * s) < n)
                { 
                    s = produceBoothMultiSimulation(s,s,t1);               
                }
                else
                {
                    s = produceBoothMultiSimulation(s,s,t1) % n;
                    t1 += MOD;
                }
                t1 += ASSIGN;
                
                //if the bit is set do a multiplication
                if((guessD[j] && j <= i)||(b1 && (j == (i+1)))||(b2 && (j == (i+2))))
                {
                    //s = (s*x)%n; modulo effort is differentiated
                    if((s * x) < n)
                    {
                        s = produceBoothMultiSimulation(s,x,t1);
                    }
                    else
                    {
                        s = produceBoothMultiSimulation(s,x,t1) % n;
                        t1 += MOD;
                    }
                    t1 +=ASSIGN;
                }
            }
        }
        return t1;
    }
    
    
    
     /* 
     * One bit guess simulation
     * 
     * @param b1 First bit guess
     * @param x  Message
     * @param guessD[] Bits already guessed
     * @param i Position up to which all bits have already been guessed
     * @return the execution time
     */
    static int calcTimeSimulation_1b(bool b, bigint x, bool guessD[], int i)
    {
        bigint n = VALUE_N;//modulo
        bigint s = x;
        
        int t1 = 0;//variable for the result
        int timeS =0;//variable for the intermediate time caused by s
        int timeX = 0;//variable for the intermediate time caused by x
        
        
        if(VARIANT == 1)
        {
            timeX = Utils::highestSigBit(x);
            
            //calculation and previous guesses and guess bit time simulation
            for(int j = 1; j <= i+1; j++)
            {
                s = (s * s) % n;
                
                t1 += timeX;
                t1 += ASSIGN;
                
                //if the bit is set do a multiplication
                if(guessD[j]||(b && (j == (i+1))))
                {
                    s = (s * x) % n;
                    
                    t1 += timeX;
                    t1 += ASSIGN;
                }
            }
        }else if(VARIANT == 2)
        {
            //calculation and previous guesses and guess bit time simulation
            for(int j = 1; j <= i+1; j++)
            {
                timeS = Utils::highestSigBit(s);
                
                //s = (s*s)%n; modulo effort is differentiated
                if((s * s) < n)
                {
                    s = s * s;
                    t1 +=(timeS * 2);
                }
                else
                {
                    s = (s * s) % n;
                    t1 += (timeS * 2);
                    t1 += MOD;
                }
                t1 += ASSIGN;
                 
                //if the bit is set do a multiplication
                if(guessD[j]||(b && (j == (i+1))))
                {
                    timeX= Utils::highestSigBit(x);
                    timeS = Utils::highestSigBit(s);
                    
                    //s = (s*x)%n; modulo effort is differentiated
                    if((s * x) < n)
                    {
                        s = s * x;
                        
                        t1 += (timeS + timeX);
                    }
                    else
                    {
                        s = (s * x) % n;
                        
                        t1 += (timeS + timeX);
                        t1 += MOD;
                    }
                    t1 += ASSIGN;
                }
            }
        }
        else if(VARIANT == 3)
        {
            //calculation and previous guesses and guess bit time simulation
            for(int j = 1; j <= i+1; j++)
            {
                timeS = Utils::numberBits(s);
                
                //s = (s*s)%n; modulo effort is differentiated
                if((s * s) < n)
                {
                    s = s * s;
                    
                    t1 += timeS * 2;
                }
                else
                {
                    s = (s * s) % n;
                    
                    t1 += timeS * 2;
                    t1 += MOD;
                }
                t1 += ASSIGN;
                
                //if the bit is set do a multiplication
                if(guessD[j]||(b && (j == (i+1))))
                {
                    timeX= Utils::numberBits(x);
                    timeS = Utils::numberBits(s);
                    
                    //s = (s*x)%n; modulo effort is differentiated
                    if((s * x) < n)
                    {
                        s = s * x;
                        t1 += timeS + timeX;
                    }
                    else
                    {
                        s = (s * x) % n;
                        t1 += timeS + timeX;
                        t1 += MOD;
                    }
                    t1 += ASSIGN;
                }
            }
        }
        return t1;
    }
};

#endif
