/*****************************************************************************
 *
 * Copyright (c) 2020, Marcel Schoppmeier
 * Westfälische Wilhelms-Universität Münster, AG Embedded Systems 
 * All rights reserved.
 *
 * This file is part of SysCSecT
 *
 * SysCSecT is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free
 * Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 *
 * SysCSecT is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SysCSecT.  If not, see <http://www.gnu.org/licenses/>.
 *
 *
 *  Please report any problems or bugs to: sysc.sec@uni-muenster.de
 *
 ****************************************************************************/

#ifndef TIMINGANALYSER
#define TIMINGANALYSER

#include "constants.h"
#include "Utils.cpp"
#include "Simulations.cpp"
#include <fstream>
#include <string>

/*
 * The timeanalyzer sends messages to the subject, measures the elapsed time per message and performs the timing analysis.
 * 
 * @param name name of the timeanalyzer
 * @param numMessages number of messages to be sent per guess run
 */
int dValue = 0;//private key in RSA

SC_MODULE(Timeanalyser)   
{
    //attributes
    sc_fifo_in<bigint> in;//port fpr sc_fifo channel
    sc_fifo_out<bigint> out;//port fpr sc_fifo channel
    sc_in<bool> cipherProceeded_in;//port for sc_signal
    int numMessages= 0;//number of messages
    int failsCount = 0;//Counter for incorrectly guessed keys 
    
    //main function
    inline void timeanalyser_main() {
        
        wait(in->data_written_event());//waits for the start signal from the subject
        
        
        //create file for results
        std::string mainPath = std::to_string(VARIANT) + "_GB" + std::to_string(GUESSBITS) + std::string("_Loop") + std::to_string(NUMBER_OF_MESSAGES) +std::string("_toNUM_") + std::to_string(TO_NUMBER);
        
        std::string pathResultChartData = RESULTPATH + std::string("/data_V") + mainPath  + std::string(".txt");
        std::string pathFails = RESULTPATH + std::string("/fails_V") + mainPath + std::string(".tex");
        
        //output file streams
        std::ofstream file; //data if key guess was correct or not
        std::ofstream fails; //keys where the guess was false
        
        //open files
        file.open(pathResultChartData);
        fails.open(pathFails);
        
        //private keys to be guessed
        for(dValue = FROM_NUMBER; dValue <= TO_NUMBER; dValue++)
        {
            //Number of attempts to guess a private key
            for(int noe = 0; noe < NUMBER_OF_EXECUTIONS; noe++)
            {
                bigint d = dValue; //current private key
                //create array to store guess bits
                int numberBitsD = Utils::highestSigBit(d); 
                bool guessD [numberBitsD] = {false};
                guessD[0] = true;//First bit is every time true
                
                //number of bits which have to be guessed
                for(int i = 0; i < numberBitsD - 1; i++)//Verschiebung um eins nach links, damit die Befuellung von simTimeValues stimmt. Bei i = 4 wird die i = 5te Bitstelle (6te durch Verschiebung) geraten
                {
                    //Init of the arrays for the simulated real time
                    sc_time t0(0,SC_PS);
                    sc_time realTimeValues [numMessages] = {t0};
                    
                    /*
                     * Init of the arrays for the calculated simulation time by the timeanlyze.r
                     * For two bit guess. One bit guess uses 00,11 only.
                     */
                    int simTimeValues00 [numMessages] = {0};
                    int simTimeValues01 [numMessages] = {0};
                    int simTimeValues10 [numMessages] = {0};
                    int simTimeValues11 [numMessages] = {0};
                    
                    
                    for(int j = 0; j < numMessages; j++) {
                        
                        //1. measures the simulated real time
                        
                        //generates random number and measures the elapsed time per processed message
                        bigint rdmMsg = Utils::generateNumber();
                        
                        sc_time t_old= sc_time_stamp();
                        out->write(rdmMsg);//sends the number to the subject
                        wait();//interrupts thread until sensitive list changes
                        
                        sc_time duration = (sc_time_stamp() - t_old);//calculates elapsed time
                        realTimeValues[j] = duration;
                        
                        
                        //2. calculates the simulation time
                        if(GUESSBITS == 1)
                        {
                            simTimeValues00[j] = Simulations::calcTimeSimulation_1b(false, rdmMsg, guessD, i);
                            simTimeValues11[j] = Simulations::calcTimeSimulation_1b(true, rdmMsg, guessD, i);
                        }
                        else if(GUESSBITS == 2)
                        {
                            simTimeValues00[j] = Simulations::calcTimeSimulation_2b(false, false, rdmMsg, d, guessD, i);
                            simTimeValues01[j] = Simulations::calcTimeSimulation_2b(false, true, rdmMsg, d, guessD, i);
                            simTimeValues10[j] = Simulations::calcTimeSimulation_2b(true, false, rdmMsg, d, guessD, i);
                            simTimeValues11[j] = Simulations::calcTimeSimulation_2b(true, true, rdmMsg, d, guessD, i);
                        }
                        
                    }
                    
                    //3. Evaluations: Calculates variance to decide which guess we must take
                    
                    if(GUESSBITS == 1)
                    {
                        int variance0 = Utils::variance(realTimeValues, simTimeValues00, numMessages);
                        int variance1 = Utils::variance(realTimeValues, simTimeValues11, numMessages);
                        
                        //Debug:
                        //cout<<"var0="<<variance0<<endl;
                        //cout<<"var1="<<variance1<<endl;
                        
                        //query to decide which bit to guess
                        guessD[i+1] = ((variance0 <= variance1) ? false : true);
                    }
                    else if(GUESSBITS == 2)
                    {
                        int variance00 = Utils::variance(realTimeValues, simTimeValues00, numMessages);
                        int variance01 = Utils::variance(realTimeValues, simTimeValues01, numMessages);
                        int variance10 = Utils::variance(realTimeValues, simTimeValues10, numMessages);
                        int variance11 = Utils::variance(realTimeValues, simTimeValues11, numMessages);
                        
                        //query to decide which bit to guess
                        guessD[i+1] = (((variance00+variance01) <= (variance10+variance11)) ? false : true);
                    }
                    
                    //Debug (Print Data)
                    /*Utils::printArray(realTimeValues, numMessages);
                     *Utils::printArray(simTimeValues00, numMessages);
                     *Utils::printArray(simTimeValues01, numMessages);
                     *Utils::printArray(simTimeValues10, numMessages);
                     *Utils::printArray(simTimeValues11, numMessages);*/
                }
                
                //Prints out whether the result is correct or not
                if( Utils::compareBinInt(guessD, numberBitsD, dValue))
                {
                    //Debug
                    //Utils::printArray(guessD,numberBitsD);
                    //cout<<"CORRECT"<<endl;
                    
                    cout<<dValue<<" "<<1<<endl;
                    file<<dValue<<" "<<1<<endl;
                }
                else
                {
                    //Debug
                    // Utils::printArray(guessD,numberBitsD);
                    // cout<<"---------------WRONG-----------------"<<endl;
                    
                    cout<<dValue<<" "<< 0 <<endl;
                    file<<dValue<<" "<< 0 <<endl;
                    
                    //write wrong guessed key into the file fails
                    fails<<std::to_string(dValue)<<":[";
                    for(int l = 0; l < numberBitsD; ++l)
                        fails<<guessD[l]<<",";
                    fails<<"]\n";
                    
                    failsCount++;//Count the wrong keys
                }
            }
        }
        fails<< "Number of fails: " << failsCount <<endl;
        
        file.close();
        fails.close();
        sc_stop();
    }
    
    //constructor
    SC_HAS_PROCESS(Timeanalyser);
    Timeanalyser(sc_module_name name, int numMessages): sc_module(name), cipherProceeded_in("cPi"), in("in"), out("out"){
        this->numMessages = numMessages;
        SC_THREAD(timeanalyser_main);
        sensitive << cipherProceeded_in.pos() << cipherProceeded_in.neg();
    }
};

#endif
