/*****************************************************************************
 *
 * Copyright (c) 2020, Marcel Schoppmeier
 * Westfälische Wilhelms-Universität Münster, AG Embedded Systems 
 * All rights reserved.
 *
 * This file is part of SysCSecT
 *
 * SysCSecT is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free
 * Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 *
 * SysCSecT is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SysCSecT.  If not, see <http://www.gnu.org/licenses/>.
 *
 *
 *  Please report any problems or bugs to: sysc.sec@uni-muenster.de
 *
 ****************************************************************************/

#ifndef UTILS_CPP
#define UTILS_CPP

#include <random>
#include <chrono>
#include <cmath>
#include <cstdio>
#include "constants.h"

/*
 * Class that encapsulates many small operations that are used in modules.
 */

class Utils{
    
private:
    
public:
    /* 
     * Generates random number
     * 
     * @return a random number between 0 and MAXBITINT.
     */
    static bigint generateNumber(){
        int time = std::chrono::system_clock::now().time_since_epoch().count(); 
        std::default_random_engine generator (time);
        std::uniform_int_distribution<int> dist (0, MAXBITINT);
        bigint rdm = dist(generator);
        
        return rdm;
    }
    
    /* 
     * Location of the first significant bit (position 0 included)
     * 
     * @param number
     * @return position of the highest significant bit
     */
    static int highestSigBit(bigint number)
    {
        int result = 0;
        
        if(number != 0)
        {
            result = (log(number.to_int())/log(2))+1;
        }
        
        return result;
    }
    
    /* 
     * Calculates the number of all bits 
     * 
     * @param number
     * @return number of bits set
     */
    static int numberBits(bigint number)
    {
        int result = 0;
        
        for(int i = 0; i < MAXBITS; ++i)
        {
            if(number[i] == 1)
                result++;
        }
        return result;
    }
    
    
    /*
     * For the Variant double, the variance declaration must also be adapted in Timeanalyser.c
     */
    /* 
     * Calculates the mean of the difference of two arrays
     * 
     * @param a First array
     * @param b Second array
     * @return mean value
     */
    /*static double mean(sc_time a[], int b[], int length)
    {
        double result = 0;
        
        for(int i = 0; i < length; ++i)
        {
            int tmp = a[i].value();
            result+= (tmp - b[i])/(double)length;
        }
        
        return result;
    }*/
    
    /* 
     * Calculates the variance of the difference of two arrays
     * 
     * @param a First array
     * @param b Second array
     * @return variance value
     */
    /*static double variance(sc_time a[], int b[], int length)
    {
        double result = 0;
        double meanAB = mean(a,b,length);
        
        for(int i = 0; i < length; ++i)
        {
            int tmp = a[i].value();
            result+= pow((tmp - b[i])- meanAB, 2)/(double)length;
        }
        
        return result;
    }*/
    
    /* 
     * Calculates the mean of the difference of two arrays
     * 
     * @param a First array
     * @param b Second array
     * @return mean value
     */
    static int mean(sc_time a[], int b[], int length)
    {
        long long int result = 0;
        
        for(int i = 0; i < length; ++i)
        {
            auto tmp = a[i].value();
            result += (tmp - b[i]);
        }
        
        result /= length;
        
        return result;
    }
    
    /* 
     * Calculates the variance of the difference of two arrays
     * 
     * @param a First array
     * @param b Second array
     * @return variance value
     */
    static int variance(sc_time a[], int b[], int length)
    {
        long long int result = 0;
        auto meanAB = mean(a, b, length);
        
        for(int i = 0; i < length; ++i)
        {
            auto tmp = a[i].value();
            auto factor = (tmp - b[i])- meanAB;
            result += factor * factor;
        }
        
        result /= length;
        
        return result;
    }
    
    /* 
     * Print the array to the command line
     * 
     * @param arr[] Array
     * @param length Length of the array
     */
    template<typename T>
    static void printArray(T arr[], int length)
    {
        for(int i = 0; i < length; ++i)
            cout<<arr[i]<< "; ";
        cout<<endl;
    }
    
    /*
     * Compares bool array with int for equality
     * 
     * @param b Bool array
     * @param length Size of the bool array
     * @param x Integer value
     * @return logical value if they are equal or not
     */
    static bool compareBinInt(bool b[], int length, int x)
    {
        int result = 0;
        
        for(int i = 0; i < length; ++i)
        {
            result+= b[i]*pow(2,length - 1 - i);
        }
        
        return (result == x);
    }
    
    /*
     * Checks whether the parameters are invalid
     * 
     * @return true if parameters are invalid
     */
    static bool invalidParameters()
    {
        //Maxbits range
        if(MAXBITS<=2 || MAXBITS > 32)
        {
            cout<<"MAXBITS parameter not valid. It is not in the interval [2,32]"<<endl;
            return 1;
        }
        
        //negative check
        if(VALUE_N<0||VALUE_N<0||NUMBER_OF_MESSAGES<0||FROM_NUMBER<0||FROM_NUMBER<0||TO_NUMBER<0)
        {
            cout<<"At least one parameter is negative. This is an invalid entry."<<endl;
            return 1;
        }
        
        //guessbits range
        if(GUESSBITS!=1 && GUESSBITS!=2)
        {
            cout<<"GUESSBITS parameter not valid. It is not in the interval [1,2]"<<endl;
            return 1;
        }
        
        //variant
        if(!((GUESSBITS==1 && (VARIANT==1 ||VARIANT==2 ||VARIANT==3))||(GUESSBITS==2 && (VARIANT==2 ||VARIANT==3 ||VARIANT==4))))
        {
            cout<<"VARIANT "<<VARIANT<<" is not valid with GUESSBITS "<<GUESSBITS<<endl;
            return 1;
        }
        
        return 0;
    }
};
#endif
